package mvd.jester.tests;

import java.math.RoundingMode;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import com.google.common.math.LongMath;
import mvd.jester.info.SchedulingInfo;
import mvd.jester.info.TerminationInfo;
import mvd.jester.info.TerminationInfo.Level;
import mvd.jester.model.Segment;
import mvd.jester.model.SortedTaskSet;
import mvd.jester.model.SynchronousTask;
import mvd.jester.priority.PriorityManager;
import mvd.jester.priority.RateMonotonic;

/**
 * MaiaBertogna
 */
public class MaiaBertogna extends AbstractTest<SynchronousTask> {

    private final Map<SynchronousTask, TerminationInfo> responseTimes;
    private final PriorityManager priorityManager;

    public MaiaBertogna(long numberOfProcessors) {
        super(numberOfProcessors);
        this.responseTimes = new HashMap<>();
        this.priorityManager = new RateMonotonic();
    }

    @Override
    public PriorityManager getPriorityManager() {
        return priorityManager;
    }

    @Override
    public SchedulingInfo runSchedulabilityCheck(SortedTaskSet<SynchronousTask> tasks) {
        responseTimes.clear();
        for (SynchronousTask t : tasks) {
            Level taskLevel = tasks.headSet(t).size() <= tasks.size() / 2 ? Level.HIGH : Level.LOW;
            long responseTime = calculateResponseTime(tasks, t);
            responseTimes.put(t, new TerminationInfo(t.getDeadline(), responseTime, taskLevel));
        }

        return new SchedulingInfo(new HashSet<>(responseTimes.values()),
                tasks.getParallelTaskRatio(), tasks.getUtilization());
    }

    @Override
    public String getName() {
        return "MaiaBertogna";
    }

    private long calculateResponseTime(Set<SynchronousTask> tasks, SynchronousTask task) {
        long minimumWcet = getMinimumWcet(task);
        long responseTime = minimumWcet;
        long previousResponseTime = 0;

        do {
            previousResponseTime = responseTime;
            long taskInterference = 0;

            for (SynchronousTask t : tasks) {
                if (t.getPeriod() < task.getPeriod()) {
                    long maxNumberOfJobsOfT = t.getMaximumParallelism();
                    for (int p = 0; p < maxNumberOfJobsOfT; ++p) {
                        taskInterference += Math.min(getTaskInterference(t, responseTime, p + 1),
                                responseTime - minimumWcet + 1);
                    }
                }
            }

            long selfInterference = 0;

            long maxNumberOfJobs = task.getMaximumParallelism();
            for (int p = 0; p < maxNumberOfJobs; ++p) {
                selfInterference +=
                        Math.min(getSelfInterference(task, p + 1), responseTime - minimumWcet + 1);
            }

            long totalInterference = LongMath.divide(taskInterference + selfInterference,
                    numberOfProcessors, RoundingMode.FLOOR);

            responseTime = minimumWcet + totalInterference;
        } while (previousResponseTime != responseTime);


        return responseTime;
    }

    private long getSelfInterference(SynchronousTask task, long parallelism) {
        long interference = 0;

        for (Segment s : task.getWorkloadDistribution()) {
            if (s.getNumberOfJobs() >= parallelism + 1) {
                interference += s.getJobWcet();
            }
        }

        return interference;
    }

    private long getTaskInterference(SynchronousTask task, long interval, long parallelism) {
        if (responseTimes.containsKey(task)) {
            long responseTime = responseTimes.get(task).getResponseTime();
            long minWcet = getMinimumWcet(task);
            long period = task.getPeriod();
            long numberOfJobs =
                    (LongMath.divide(interval + responseTime - minWcet, period, RoundingMode.FLOOR)
                            + 1);

            long workload = 0;

            for (Segment s : task.getWorkloadDistribution()) {
                if (s.getNumberOfJobs() >= parallelism) {
                    workload += s.getJobWcet();
                }
            }

            long interference = numberOfJobs * workload;

            return interference;
        } else {
            throw new RuntimeException("Task was not found in task set!");
        }
    }

    private long getMinimumWcet(SynchronousTask task) {
        long minWcet = 0;
        for (Segment s : task.getWorkloadDistribution()) {
            minWcet += s.getJobWcet();
        }

        return minWcet;
    }

}
