package mvd.jester.model;

import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.mock;
import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.Set;
import java.util.concurrent.ThreadLocalRandom;
import javax.imageio.ImageIO;
import com.google.common.math.DoubleMath;
import com.mxgraph.layout.mxCircleLayout;
import com.mxgraph.layout.mxIGraphLayout;
import com.mxgraph.util.mxCellRenderer;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.ext.JGraphXAdapter;
import org.jgrapht.graph.DefaultEdge;
import org.junit.Rule;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.rules.TemporaryFolder;
import mvd.jester.model.SystemSetup.DagTaskBuilder;


public class TestSystemSetup {
    @Rule
    public TemporaryFolder tf = new TemporaryFolder();

    private static final int NUMBER_OF_RUNS = 1000;

    @Test
    @DisplayName("Check if randomly generated task parameters lie within the specified ranges.")
    public void testRandomTaskSetGeneration() {

        for (int i = 0; i < NUMBER_OF_RUNS; ++i) {
            long numberOfProcessors = ThreadLocalRandom.current().nextLong(2, 8);
            SystemSetup.SynchronousTaskBuilder systemSetupBuilder =
                    new SystemSetup.SynchronousTaskBuilder()//
                            .setNumberOfSegments(1, 7)//
                            .setNumberOfJobs(2, 10)//
                            .setPeriods(100, 1000, 1000)//
                            .setNumberOfProcessors(numberOfProcessors);

            Set<SynchronousTask> taskSet = systemSetupBuilder.generateTaskSet();

            // assertTrue(systemSetup.getNumberOfProcessors() == numberOfProcessors);

            for (SynchronousTask t : taskSet) {
                assertTrue(t.getPeriod() >= 100);
                assertTrue(t.getPeriod() <= 1000);
                assertTrue(t.getDeadline() == t.getPeriod());

                long maxJobWcet = t.getPeriod() / t.getWorkloadDistribution().size();

                for (Segment s : t.getWorkloadDistribution()) {
                    assertTrue(s.getJobWcet() >= 1);
                    assertTrue(s.getJobWcet() <= maxJobWcet);
                    assertTrue(s.getNumberOfJobs() >= 1);
                    assertTrue(s.getNumberOfJobs() <= 10);
                }

                assertTrue(t.getWorkloadDistribution().size() >= 1);
                assertTrue(t.getWorkloadDistribution().size() <= 7);

            }

            assertTrue(systemSetupBuilder.addTask(taskSet));
        }
    }

    @Test
    @DisplayName("Check Getters and Setters.")
    void testGettersAndSetters() {
        @SuppressWarnings("unchecked")
        Set<SynchronousTask> t1 = mock(Set.class);

        @SuppressWarnings("unchecked")
        Set<SynchronousTask> t2 = mock(Set.class);

        SystemSetup<SynchronousTask> systemSetup = new SystemSetup<>(t1, 2);

        systemSetup.setTasks(t2);
        assertTrue(systemSetup.getTasks() == t2);
    }

    @Test
    @DisplayName("Check if DagTaskBuilder works correctly")
    void testDagTaskBuilder() {
        DagTaskBuilder builder = new DagTaskBuilder();
        DirectedAcyclicGraph<Job, DefaultEdge> j = builder.generateTask().getJobDag();

        JGraphXAdapter<Job, DefaultEdge> graphAdapter = new JGraphXAdapter<Job, DefaultEdge>(j);

        mxIGraphLayout layout = new mxCircleLayout(graphAdapter);
        layout.execute(graphAdapter.getDefaultParent());

        BufferedImage image =
                mxCellRenderer.createBufferedImage(graphAdapter, null, 2, Color.WHITE, true, null);

        File imgFile = new File("src/test/resources/graph.png");
        try {
            ImageIO.write(image, "PNG", imgFile);
        } catch (Exception e) {
        }

        assertTrue(imgFile.exists());

    }


    @Test
    @DisplayName("Check if utilization works correctly.")
    void testUtil() {
        for (double d = 0.25; d < 4; d += 0.25) {
            DagTaskBuilder builder = new DagTaskBuilder();
            Set<DagTask> taskSet = builder.generateTaskSet(d);
            double taskSetUtil = 0;
            for (DagTask t : taskSet) {
                taskSetUtil += t.getUtilization();
            }

            assertTrue(DoubleMath.fuzzyEquals(taskSetUtil, d, 0.002));
        }
    }

    // @Test
    // @DisplayName("Check if parser works correclty.")
    // void testParser() throws IOException {
    // SystemSetup systemSetup = new SystemSetup.Builder().setNumberOfSegments(1, 7)
    // .setNumberOfJobs(2, 10).setPeriods(100, 1000, 1000).build();

    // systemSetup.writeToFile(null);
    // systemSetup.writeToFile(tf.getRoot().getAbsolutePath());
    // }
}
